@@ -1,7 +1,8 @@
 //! Tokay main executable
 
-use ::tokay::compiler::Compiler;
-use ::tokay::repl::{repl, Stream};
+use tokay::compiler::Compiler;
+use tokay::repl::{repl, Stream};
+use tokay::Object;
 
 #[macro_use]
 extern crate clap;
@@ -79,15 +80,25 @@ fn main() {
         }
     }
 
-    // In case no stream is specified, use empty string as default input stream.
-    if streams.len() == 0 {
-        streams.push(("", RefCell::new(Stream::String("".to_string()))));
-    }
-
     if let Some(mut program) = program {
         let mut compiler = Compiler::new();
 
         if let Ok(program) = compiler.compile(program.get_reader()) {
+            // In case no stream but a program is specified, use stdin as input stream.
+            if streams.len() == 0 {
+                streams.push((
+                    "",
+                    // When program's main is consuming, read from stdin
+                    if program.main().is_consuming() {
+                        RefCell::new(Stream::Stdin)
+                    }
+                    // otherwise just work on an empty input
+                    else {
+                        RefCell::new(Stream::String("".to_string()))
+                    },
+                ));
+            }
+
             for (name, stream) in &streams {
                 let ret = program.run_from_reader(stream.borrow_mut().get_reader());
 
@@ -101,13 +112,19 @@ fn main() {
                             print!("\n")
                         }
                     }
-                    Ok(Some(value)) => println!("{}", value),
+                    Ok(Some(value)) => println!("{}", value.to_string()),
                     Err(error) => eprintln!("{}", error),
                 }
             }
         }
     } else {
         print_version();
+
+        // In case no stream was specified and REPL fires up, use empty string as input stream.
+        if streams.len() == 0 {
+            streams.push(("", RefCell::new(Stream::String("".to_string()))));
+        }
+
         repl(streams);
     }
 }
