@@ -9,8 +9,11 @@ use std::fs::File;
 use std::io::Read;
 use std::process::Command;
 
-// Universal function to run test-cases from files with expected output.
-fn run_testcase(filename: &'static str) {
+/** Test case utility function to run test-cases from files with expected output.
+
+This function currently requires that a tokay debug executable is compiled before test cases are run.
+*/
+pub(crate) fn testcase(filename: &'static str) {
     let mut code = String::new();
 
     match File::open(filename) {
@@ -99,7 +102,7 @@ fn run_testcase(filename: &'static str) {
 // Test for literals
 fn literal() {
     assert_eq!(
-        compile_and_run(
+        run(
             "\
             1337 \
             23.5 \
@@ -118,7 +121,7 @@ fn literal() {
 fn expression() {
     // Simple Integer expressions
     assert_eq!(
-        compile_and_run(
+        run(
             "\
             1 + 2 \
             100 - 99 \
@@ -133,7 +136,7 @@ fn expression() {
 
     // Simple Float expressions
     assert_eq!(
-        compile_and_run(
+        run(
             "\
             .2  + .3 \
             1.9 - 1.1 \
@@ -154,7 +157,7 @@ fn expression() {
 
     // Simple String expressions
     assert_eq!(
-        compile_and_run(
+        run(
             "\
             \"a\" + \"b\" \
             \"a\" + 2 \
@@ -177,7 +180,7 @@ fn expression() {
 
     // Equality Comparisons
     assert_eq!(
-        compile_and_run(
+        run(
             "\
             x = 42 \
             42 == 42 \
@@ -207,7 +210,7 @@ fn expression() {
 
     // Ordered Comparisons
     assert_eq!(
-        compile_and_run(
+        run(
             "\
             42 >= 42 \
             42 <= 42 \
@@ -233,7 +236,7 @@ fn expression() {
 
     // Logical AND and OR
     assert_eq!(
-        compile_and_run(
+        run(
             "\
             a = true
             b = false
@@ -254,7 +257,7 @@ fn expression() {
 
     // Unary operations
     assert_eq!(
-        compile_and_run(
+        run(
             "\
             -1 \
             (-(-1)) \
@@ -272,7 +275,7 @@ fn expression() {
 fn operations() {
     // Test assignment-operations
     assert_eq!(
-        compile_and_run(
+        run(
             "
             a = true a \
             a + 2 == 3 \
@@ -291,7 +294,7 @@ fn operations() {
     // Tests for pre- and post-increment and -decrements
     // These require spaces in some situations to find correct path throug meaning
     assert_eq!(
-        compile_and_run(
+        run(
             "
             a = 1 \
             a \
@@ -306,6 +309,44 @@ fn operations() {
         ),
         Ok(Some(value![[1, 1, 3, 3, 4, 12, 2, 8, 3]]))
     );
+
+    // Inline add int to itself
+    assert_eq!(
+        run(
+            "
+            i = 2 \
+            i *= i \
+            i
+            ",
+            ""
+        ),
+        Ok(Some(value!(4)))
+    );
+
+    // Inline add list to itself
+    // fixme: Should be put into the specific modules, e.g. value/list.rs
+    assert_eq!(
+        run(
+            "
+            a = (1,2) \
+            b = (3,4) \
+            a + b \
+            a \
+            b \
+            a += b \
+            a \
+            b
+            ",
+            ""
+        ),
+        Ok(Some(value!([
+            [1, 2, 3, 4],
+            [1, 2],
+            [3, 4],
+            [1, 2, 3, 4],
+            [3, 4]
+        ])))
+    );
 }
 
 #[test]
@@ -313,7 +354,7 @@ fn operations() {
 fn variables() {
     // Test store-hold-global
     assert_eq!(
-        compile_and_run(
+        run(
             "
             a = b = 10
             a++
@@ -326,7 +367,7 @@ fn variables() {
 
     // Test store-hold-local
     assert_eq!(
-        compile_and_run(
+        run(
             "
             f : @{
                 a = b = 10
@@ -343,7 +384,7 @@ fn variables() {
 
     // Test store-hold-capture
     assert_eq!(
-        compile_and_run(
+        run(
             "
             10 20 $1 = $2 = 30 ++$1 $2
             ",
@@ -354,7 +395,7 @@ fn variables() {
 
     // Test store-hold-aliased-capture
     assert_eq!(
-        compile_and_run(
+        run(
             "
             a => 10 b => 20 $a = $b = 30 c => ++$a d => $b
             ",
@@ -368,7 +409,7 @@ fn variables() {
 // Test-case for scoping
 fn scoping() {
     assert_eq!(
-        compile_and_run(include_str!("../tests/test_scopes.tok"), ""),
+        run(include_str!("../tests/test_scopes.tok"), ""),
         Ok(Some(value![[10, 2000, 1072]]))
     );
 }
@@ -380,7 +421,7 @@ fn scoping() {
 fn collections() {
     // Lists
     assert_eq!(
-        compile_and_run(
+        run(
             "
             (1 2 3) \
             (1, 2, 3) \
@@ -397,7 +438,7 @@ fn collections() {
 
     // Dicts
     assert_eq!(
-        compile_and_run(
+        run(
             "
             x = 10
             (a => 1 b => 2 c => 3) \
@@ -423,25 +464,22 @@ fn token_modifiers() {
 
     // Simple touch
     assert_eq!(
-        compile_and_run("'a' 'b'", s),
+        run("'a' 'b'", s),
         Ok(Some(value![[["a", "b"], ["a", "b"]]]))
     );
 
     // Touch and match
-    assert_eq!(
-        compile_and_run("'a' ''b''", s),
-        Ok(Some(value![["b", "b"]]))
-    );
+    assert_eq!(run("'a' ''b''", s), Ok(Some(value![["b", "b"]])));
 
     // Match with positive modifier
     assert_eq!(
-        compile_and_run("'a' ''b''+", s),
+        run("'a' ''b''+", s),
         Ok(Some(value![["b", ["b", "b", "b"]]]))
     );
 
     // Match with kleene and positive modifiers
     assert_eq!(
-        compile_and_run("''a''* ''b''+", s),
+        run("''a''* ''b''+", s),
         Ok(Some(value![[
             ["a", "b"],
             ["a", ["b", "b", "b"]],
@@ -451,203 +489,60 @@ fn token_modifiers() {
 
     // Touch with kleene and positive modifiers
     assert_eq!(
-        compile_and_run("'a'* ''b''+", s),
+        run("'a'* ''b''+", s),
         Ok(Some(value![["b", ["b", "b", "b"], ["b", "b"]]]))
     );
 
     // Character classes
-    assert_eq!(
-        compile_and_run("[a-z]", &s[..2]),
-        Ok(Some(value![["a", "b"]]))
-    );
+    assert_eq!(run("[a-z]", &s[..2]), Ok(Some(value![["a", "b"]])));
 
     assert_eq!(
-        compile_and_run("[a-z]+", s),
+        run("[a-z]+", s),
         Ok(Some(value![["ab", "abbb", "bb", "def"]]))
     );
 
     assert_eq!(
-        compile_and_run("[^ ']+", s),
+        run("[^ ']+", s),
         Ok(Some(value![[
             "ab", "abbb", "bb", "123", "ABC", "456", "def"
         ]]))
     );
 
     // Built-in token
-    assert_eq!(compile_and_run("Integer", s), Ok(Some(value![[123, 456]])));
+    assert_eq!(run("Int", s), Ok(Some(value![[123, 456]])));
 
     // Parsing with sequences and modifiers
 
     assert_eq!(
-        compile_and_run("''a'' {''b'' ''c''}* ''d''", "abcbcd"),
+        run("''a'' {''b'' ''c''}* ''d''", "abcbcd"),
         Ok(Some(value![["a", [["b", "c"], ["b", "c"]], "d"]]))
     );
 
     assert_eq!(
-        compile_and_run("''a'' {''b'' ''c''}+ ''d''", "abcbcd"),
+        run("''a'' {''b'' ''c''}+ ''d''", "abcbcd"),
         Ok(Some(value![["a", [["b", "c"], ["b", "c"]], "d"]]))
     );
 
     assert_eq!(
-        compile_and_run("''a'' {''b'' ''c''}* ''d''", "ad"),
+        run("''a'' {''b'' ''c''}* ''d''", "ad"),
         Ok(Some(value![["a", "d"]]))
     );
 
-    assert_eq!(
-        compile_and_run("''a'' {''b'' ''c''}+ ''d''", "ad"),
-        Ok(None)
-    );
+    assert_eq!(run("''a'' {''b'' ''c''}+ ''d''", "ad"), Ok(None));
 
     assert_eq!(
-        compile_and_run("{ Word { ',' _ }? }+", "Hello,   World,  Beta,  Test"),
+        run("{ Word { ',' _ }? }+", "Hello,   World,  Beta,  Test"),
         Ok(Some(value![["Hello", "World", "Beta", "Test"]]))
     );
 
     // todo: more token tests, please!
 }
 
-#[test]
-// Test for built-in tokens
-fn builtin_tokens() {
-    let gliders = "Glasflügel Libelle 201b\tG102 Astir  \nVentus_2cT";
-
-    assert_eq!(
-        compile_and_run("Identifier", gliders),
-        Ok(Some(value!([
-            "Glasflügel",
-            "Libelle",
-            "b",
-            "G102",
-            "Astir",
-            "Ventus_2cT"
-        ])))
-    );
-
-    assert_eq!(
-        compile_and_run("Integer", gliders),
-        Ok(Some(value!([201, 102, 2])))
-    );
-
-    assert_eq!(
-        compile_and_run("Whitespaces", gliders),
-        Ok(Some(value!([" ", " ", "\t", " ", "  \n"])))
-    );
-
-    assert_eq!(
-        compile_and_run("Word", gliders),
-        Ok(Some(value!([
-            "Glasflügel",
-            "Libelle",
-            "b",
-            "G",
-            "Astir",
-            "Ventus",
-            "cT"
-        ])))
-    );
-
-    // Builtin whitespace handling
-    let abc = "abc   \tdef  abcabc= ghi abcdef";
-
-    assert_eq!(
-        compile_and_run("Word _; ", abc),
-        Ok(Some(value![["abc", "def", "abcabc", "ghi", "abcdef"]]))
-    );
-
-    assert_eq!(
-        compile_and_run("Word __; ", abc),
-        Ok(Some(value![["abc", "def", "ghi"]]))
-    );
-}
-
-// Tests for parselets ----------------------------------------------------------------------------
-
-#[test]
-// Testing static function with arguments
-fn parselet_static_with_args() {
-    assert_eq!(
-        compile_and_run(
-            "
-            faculty : @x {
-                if !x return 1
-                x * faculty(x - 1)
-            }
-
-            faculty(4)
-            ",
-            ""
-        ),
-        Ok(Some(value!(24)))
-    );
-}
-
-#[test]
-// Testing variable function with arguments
-fn parselet_variable_with_args() {
-    assert_eq!(
-        compile_and_run(
-            "
-            faculty = @x {
-                if !x return 1
-                x * faculty(x - 1)
-            }
-
-            faculty(4)
-            ",
-            ""
-        ),
-        Ok(Some(value!(24)))
-    );
-}
-
-#[test]
-// Testing left-recursive parselets
-fn parselet_leftrec() {
-    assert_eq!(
-        compile_and_run("P: @{ P? ''a'' }\nP", "aaaa"),
-        Ok(Some(value!([[["a", "a"], "a"], "a"])))
-    );
-
-    // todo: More examples here please!
-}
-
-#[test]
-// Testing compile- and run-time error reporting
-fn parselet_call_error_reporting() {
-    // Tests for calling functions with wrong parameter counts
-    for (call, msg) in [
-        ("foo()", "Line 2, column 1: Call to unresolved symbol 'foo'"),
-        (
-            "f()",
-            "Line 2, column 1: Call to 'f' doesn't accept any arguments",
-        ),
-        (
-            "f(1, 2, 3, 4)",
-            "Line 2, column 1: Too many parameters, 3 possible, 4 provided",
-        ),
-        ("f(c=10, d=3)", "Line 2, column 1: Parameter 'a' required"),
-        (
-            "f(1, c=10, d=3)",
-            "Line 2, column 1: Parameter 'd' provided to call but not used",
-        ),
-    ] {
-        let call = format!("f : @a, b=2, c {{ a b c }}\n{}", call);
-        println!("calling {:?}, expecting {:?}", call, msg);
-
-        assert_eq!(compile_and_run(&call, ""), Err(msg.to_owned()));
-    }
-}
-
-#[test]
-fn kaputt() {
-    assert_eq!(compile_and_run("{{}}", ""), Ok(Some(Value::Void)));
-}
-
 #[test]
 // Testing examples provided in the examples folder
 fn examples() {
     assert_eq!(
-        compile_and_run(
+        run(
             include_str!("../examples/planets.tok"),
             "Mercury Venus Earth Mars"
         ),
@@ -661,7 +556,7 @@ fn examples() {
 
     /*
     assert_eq!(
-        compile_and_run(
+        run(
             include_str!("../examples/planets2.tok"),
             "Mercury Venus Earth Mars Jupiter Saturn Uranus Neptune"
         ),
@@ -679,18 +574,18 @@ fn examples() {
     */
 
     assert_eq!(
-        compile_and_run(include_str!("../examples/expr.tok"), "1+2*3+4"),
+        run(include_str!("../examples/expr.tok"), "1+2*3+4"),
         Ok(Some(value!(11)))
     );
 
     // todo: Would be nice to test against stdout
     assert_eq!(
-        compile_and_run(include_str!("../examples/expr_with_ast.tok"), "1+2*3+4"),
+        run(include_str!("../examples/expr_with_ast.tok"), "1+2*3+4"),
         Ok(None)
     );
 
     assert_eq!(
-        compile_and_run(
+        run(
             include_str!("../examples/expr_with_spaces.tok"),
             "1 +  \t 2 \n *  3 + 4"
         ),
@@ -698,85 +593,22 @@ fn examples() {
     );
 
     assert_eq!(
-        compile_and_run(include_str!("../examples/faculty.tok"), ""),
+        run(include_str!("../examples/faculty.tok"), ""),
         Ok(Some(value!(24)))
     );
-}
 
-#[test]
-// Testing sequence captures
-fn capture() {
-    assert_eq!(
-        compile_and_run("'a' 'b' $1 * 2 + $2 * 3", "ab"),
-        Ok(Some(value!("aabbb")))
-    );
-
-    assert_eq!(
-        compile_and_run("a=2 'a' 'b' $(a + 1) * 3+ $(a) * 2", "ab"),
-        Ok(Some(value!("bbbaa")))
-    );
+    // todo: Move this to a separate function
+    testcase("tests/test_piped_grammar.tok");
 }
 
 // Tests for control flow -------------------------------------------------------------------------
 
-#[test]
-// Testing parselet begin and end special patterns
-fn begin_end() {
-    assert_eq!(
-        compile_and_run(
-            "
-            begin { x = 0 1337 }
-            end 1338
-
-            P: @{ 'lol' x = x + 1 x }
-            P",
-            "lolalolaalolol"
-        ),
-        Ok(Some(value!([1337, 1, 2, 3, 1338])))
-    );
-
-    assert_eq!(
-        compile_and_run(
-            "
-            begin x = 1
-
-            'lol' $1 * x x x = x + 1",
-            "lolAlolBlol"
-        ),
-        Ok(Some(value!([["lol", 1], ["lollol", 2], ["lollollol", 3]])))
-    );
-
-    // begin and end without any input
-    assert_eq!(
-        compile_and_run(
-            "
-            begin 1
-            2 3 4
-            end 5
-            ",
-            ""
-        ),
-        Ok(Some(value!([1, [2, 3, 4], 5])))
-    )
-}
-
-#[test]
-// Testing parselet repeat
-fn repeat() {
-    assert_eq!(
-        compile_and_run("P: @{ 'a' repeat $1 }\nP", "aaaa"),
-        Ok(Some(value!(["a", "a", "a", "a"])))
-    );
-
-    // todo: More examples here please!
-}
-
 #[test]
 // Testing if...else construct
 fn if_else() {
     // These expressions are optimized by the compiler
     assert_eq!(
-        compile_and_run(
+        run(
             "
             if true 1 \
             if false 2 \
@@ -789,7 +621,7 @@ fn if_else() {
 
     // These expressions are evaluated at compile time
     assert_eq!(
-        compile_and_run(
+        run(
             "
             b = true
             nb = false
@@ -803,14 +635,14 @@ fn if_else() {
         Ok(Some(value!([1, 3, 5])))
     );
 
-    run_testcase("tests/test_if.tok");
+    testcase("tests/test_if.tok");
 }
 
 #[test]
 // tests for push and next
 fn push_next() {
     assert_eq!(
-        compile_and_run(
+        run(
             "
             1 2 3 next
             4 5 6 push 7
@@ -826,147 +658,7 @@ fn push_next() {
 #[test]
 // tests for push and next
 fn loops() {
-    run_testcase("tests/test_loop.tok");
-    run_testcase("tests/test_for.tok");
-    run_testcase("tests/err_break_continue.tok");
-}
-
-// Tests for compiler behavior --------------------------------------------------------------------
-
-#[test]
-// Testing several special parsing constructs and error reporting
-fn compiler_structure() {
-    // Tests for blocks and empty blocks
-    assert_eq!(
-        compile_and_run(
-            "
-            a = {}
-            b = {
-            }
-            c = {
-                1
-                2
-                3
-            }
-
-            a b c
-            ",
-            ""
-        ),
-        Ok(Some(value!(3)))
-    );
-
-    // Test for invalid input when EOF is expected
-    assert_eq!(
-        compile_and_run("{}}", ""),
-        Err("Line 1, column 3: Parse error, expecting end-of-file".to_string())
-    );
-}
-
-#[test]
-// Tests for correct identifier names for various value types
-fn compiler_identifier_naming() {
-    run_testcase("tests/err_compiler_identifier_names.tok");
-}
-
-#[test]
-// Tests for compiler string, match and ccl escaping
-fn compiler_unescaping() {
-    assert_eq!(
-        compile_and_run(
-            "\"test\\\\yes\n\\xCA\\xFE\t\\100\\x5F\\u20ac\\U0001F98E\"",
-            ""
-        ),
-        Ok(Some(value!("test\\yes\nÊþ\t@_€🦎")))
-    );
-
-    assert_eq!(
-        compile_and_run(
-            "'hello\\nworld'", // double \ quotation required
-            "hello\nworld"
-        ),
-        Ok(Some(value!("hello\nworld")))
-    );
-
-    assert_eq!(
-        compile_and_run(
-            "[0-9\\u20ac]+", // double \ quotation required
-            "12345€ €12345"
-        ),
-        Ok(Some(value!(["12345€", "€12345"])))
-    );
-
-    assert_eq!(
-        compile_and_run(
-            "'hello\\nworld'", // double \ quotation required
-            "hello\nworld"
-        ),
-        Ok(Some(value!("hello\nworld")))
-    );
-
-    assert_eq!(
-        compile_and_run(
-            "[0-9\\u20ac]+", // double \ quotation required
-            "12345€ €12345"
-        ),
-        Ok(Some(value!(["12345€", "€12345"])))
-    );
-}
-
-// Tests for builtins -----------------------------------------------------------------------------
-
-#[test]
-// Tests for builtin functions
-fn builtins() {
-    // ord/chr
-    assert_eq!(
-        compile_and_run("i = ord(\"€\"); i chr(i)", ""),
-        Ok(Some(value![[(8364 as usize), "€"]]))
-    );
-
-    assert_eq!(
-        compile_and_run("ord(\"12\")", ""),
-        Err(
-            "Line 1, column 1: ord() expected single character, but received string of length 2"
-                .to_string()
-        )
-    );
-
-    assert_eq!(
-        compile_and_run("ord(\"\")", ""),
-        Err(
-            "Line 1, column 1: ord() expected single character, but received string of length 0"
-                .to_string()
-        )
-    );
-}
-
-#[test]
-// Tests for builtin string functions
-fn builtins_str() {
-    assert_eq!(
-        compile_and_run(
-            "
-            \"abcäöü\".upper() \
-            \"ABCÄÖÜ\".lower() \
-            \"hello world\".replace(\"l\") \
-            \"hello world\".replace(\"l\", n=2) \
-            \"hello world\".replace(\"l\", \"x\") \
-            \"hello world\".replace(\"l\", \"x\", 2) \
-            \"hello world\".replace(\"l\").upper() \
-            \"Tokay\".upper()[1] \
-            ",
-            ""
-        ),
-        Ok(Some(value![[
-            "ABCÄÖÜ",
-            "abcäöü",
-            "heo word",
-            "heo world",
-            "hexxo worxd",
-            "hexxo world",
-            "HEO WORD",
-            "O"
-        ]]))
-    );
+    testcase("tests/test_loop.tok");
+    testcase("tests/test_for.tok");
+    testcase("tests/err_break_continue.tok");
 }
